module LinearSolvers

using LinearAlgebra
using Adapt

using Printf
using SparseArrays

import Base: show

using KernelAbstractions
import Krylov

import ..ExaPF: xnorm

import Base.size, Base.sizeof, Base.format_bytes

using KLU
import Krylov
using KrylovPreconditioners

const KA = KernelAbstractions
const KP = KrylovPreconditioners

export list_solvers, default_linear_solver
export DirectSolver, Bicgstab
export do_scaling, scaling!

@enum(
    SolveStatus,
    Unsolved,
    MaxIterations,
    NotANumber,
    Converged,
    Diverged,
)

abstract type AbstractLinearSolver end
abstract type AbstractIterativeLinearSolver <: AbstractLinearSolver end

"""
    list_solvers(::KernelAbstractions.Device)

List linear solvers available on current device.

"""
function list_solvers end

"""
    ldiv!(solver, x, A, y)
    ldiv!(solver, x, y)

* `solver::AbstractLinearSolver`: linear solver to solve the system
* `x::AbstractVector`: Solution
* `A::AbstractMatrix`: Input matrix
* `y::AbstractVector`: RHS

Solve the linear system ``A x = y`` using the algorithm
specified in `solver`. If `A` is not specified, the function
will used directly the factorization stored inside `solver`.

"""
function ldiv! end

"""
    rdiv!(solver, x, A, y)
    rdiv!(solver, x, y)

* `solver::AbstractLinearSolver`: linear solver to solve the system
* `x::AbstractVector`: Solution
* `A::AbstractMatrix`: Input matrix
* `y::AbstractVector`: RHS

Solve the linear system ``A^⊤ x = y`` using the algorithm
specified in `solver`. If `A` is not specified, the function
will used directly the factorization stored inside `solver`.

"""
function rdiv! end

_get_type(J) = error("No handling of sparse Jacobian type defined in LinearSolvers")
_get_type(J::SparseMatrixCSC) = Vector{Float64}
do_scaling(linear_solver) = false
scaling!(A,b) = nothing

"""
    DirectSolver <: AbstractLinearSolver

Solve linear system ``A x = y`` with direct linear algebra.

* On the CPU, `DirectSolver` uses UMFPACK to solve the linear system
* On CUDA GPU, `DirectSolver` redirects the resolution to the method `CUSOLVER.csrlsvqr`

"""
struct DirectSolver{Fac<:Union{Nothing, LinearAlgebra.Factorization}} <: AbstractLinearSolver
    factorization::Fac
end

DirectSolver(J; options...) = DirectSolver(klu(J))
DirectSolver() = DirectSolver(nothing)

function update!(s::DirectSolver, J::AbstractMatrix)
    klu!(s.factorization, J) # Update factorization inplace
end

# Reuse factorization in update
function ldiv!(s::DirectSolver{<:LinearAlgebra.Factorization}, y::AbstractVector, J::AbstractMatrix, x::AbstractVector; options...)
    LinearAlgebra.ldiv!(y, s.factorization, x) # Forward-backward solve
    return 0
end
# Solve system Ax = y
function ldiv!(s::DirectSolver{<:LinearAlgebra.Factorization}, y::AbstractArray, x::AbstractArray; options...)
    LinearAlgebra.ldiv!(y, s.factorization, x) # Forward-backward solve
    return 0
end
function ldiv!(s::DirectSolver{<:LinearAlgebra.Factorization}, y::AbstractArray; options...)
    LinearAlgebra.ldiv!(s.factorization, y) # Forward-backward solve
    return 0
end
# Solve system A'x = y
function rdiv!(s::DirectSolver{<:LinearAlgebra.Factorization}, y::AbstractArray, x::AbstractArray)
    LinearAlgebra.ldiv!(y, s.factorization', x) # Forward-backward solve
    return 0
end

function ldiv!(::DirectSolver{Nothing}, y::Vector, J::AbstractMatrix, x::Vector; options...)
    F = lu(J)
    LinearAlgebra.ldiv!(y, F, x)
    return 0
end

update!(solver::AbstractIterativeLinearSolver, J::SparseMatrixCSC) = KP.update!(solver.precond, J)

"""
    Dqgmres <: AbstractIterativeLinearSolver
    Dqgmres(precond; verbose=false, memory=4)

Wrap `Krylov.jl`'s Dqgmres algorithm to solve iteratively the linear system
``A x = y``.
"""
struct Dqgmres <: AbstractIterativeLinearSolver
    inner::Krylov.DqgmresSolver
    precond::AbstractKrylovPreconditioner
    memory::Int
    verbose::Bool
end
function Dqgmres(J::AbstractSparseMatrix;
    P=BlockJacobiPreconditioner(J), memory=4, verbose=false
)
    n, m = size(J)
    S = _get_type(J)
    solver = Krylov.DqgmresSolver(n, m, memory, S)
    return Dqgmres(solver, P, memory, verbose)
end

function ldiv!(solver::Dqgmres,
    y::AbstractVector, J::AbstractMatrix, x::AbstractVector; options...
)
    Krylov.dqgmres!(solver.inner, J, x; N=solver.precond)
    copyto!(y, solver.inner.x)
    return length(solver.inner.stats.residuals)
end

"""
    Bicgstab <: AbstractIterativeLinearSolver
    Bicgstab(precond; verbose=0, rtol=1e-10, atol=1e-10)

Wrap `Krylov.jl`'s BICGSTAB algorithm to solve iteratively the linear system
``A x = y``.
"""
struct Bicgstab <: AbstractIterativeLinearSolver
    inner::Krylov.BicgstabSolver
    precond::AbstractKrylovPreconditioner
    verbose::Int
    atol::Float64
    rtol::Float64
    ldiv::Bool
    scaling::Bool
    maxiter::Int64
end
do_scaling(linear_solver::Bicgstab) = linear_solver.scaling
function Bicgstab(J::AbstractSparseMatrix;
    P=BlockJacobiPreconditioner(J), verbose=0, rtol=1e-10, atol=1e-10, ldiv=false, scaling=false, maxiter=size(J,1)
)
    n, m = size(J)
    S = _get_type(J)
    solver = Krylov.BicgstabSolver(n, m, S)
    return Bicgstab(solver, P, verbose, atol, rtol, ldiv, scaling, maxiter)
end

function ldiv!(solver::Bicgstab,
    y::AbstractVector, J::AbstractMatrix, x::AbstractVector;
    max_atol = solver.atol, max_rtol = solver.rtol, options...
)

    atol = max_atol < solver.atol ? solver.atol : max_atol
    rtol = max_rtol < solver.rtol ? solver.rtol : max_rtol
    Krylov.bicgstab!(
        solver.inner, J, x;
        N=solver.precond,
        atol=atol,
        rtol=rtol,
        verbose=solver.verbose,
        history=true,
        ldiv=solver.ldiv,
        itmax=solver.maxiter,
    )
    if solver.inner.stats.status == "breakdown αₖ == 0"
        @warn("BICGSTAB failed to converge. Final status is $(solver.inner.stats.status)")
    end
    copyto!(y, solver.inner.x)
    return solver.inner.stats.niter
end

"""
    list_solvers(::KA.CPU)

List all linear solvers available solving the power flow on the CPU.
"""
list_solvers(::KA.CPU) = [DirectSolver, Dqgmres, Bicgstab]

"""
    default_linear_solver(A, ::KA.CPU)

Default linear solver on the CPU.
"""
default_linear_solver(A::SparseMatrixCSC, device::KA.CPU) = DirectSolver(A)
end
